import logging

from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.generics import (
    AddRemoveView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms import SmartLinkForm
from ..icons import (
    icon_document_smart_link_instance_list, icon_document_type_smart_links,
    icon_smart_link_create, icon_smart_link_delete,
    icon_smart_link_document_type_list, icon_smart_link_edit,
    icon_smart_link_instance_detail, icon_smart_link_list,
    icon_smart_link_setup
)
from ..links import link_smart_link_create
from ..models import ResolvedSmartLink, SmartLink
from ..permissions import (
    permission_resolved_smart_link_view, permission_smart_link_create,
    permission_smart_link_delete, permission_smart_link_edit,
    permission_smart_link_view
)

logger = logging.getLogger(name=__name__)


class DocumentResolvedSmartLinkDocumentListView(
    ExternalObjectViewMixin, DocumentListView
):
    external_object_permission = permission_resolved_smart_link_view
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    view_icon = icon_smart_link_instance_detail

    def dispatch(self, request, *args, **kwargs):
        self.resolved_smart_link = self.get_resolved_smart_link()

        return super().dispatch(request, *args, **kwargs)

    def get_document_queryset(self):
        try:
            queryset = self.resolved_smart_link.get_linked_documents_for(
                document=self.external_object
            )
        except Exception as exception:
            queryset = Document.objects.none()

            # Check if the user has the smart link edit permission before
            # showing the exception text.
            try:
                AccessControlList.objects.check_access(
                    obj=self.resolved_smart_link,
                    permissions=(permission_smart_link_edit,),
                    user=self.request.user
                )
            except PermissionDenied:
                """User doesn't have the required permission."""
            else:
                messages.error(
                    message=_(
                        'Resolved smart link query error: %s' % exception
                    ), request=self.request
                )

        return queryset

    def get_extra_context(self):
        try:
            resolved_smart_link_label = self.resolved_smart_link.get_label_for(
                document=self.external_object
            )
        except Exception as exception:
            resolved_smart_link_label = self.resolved_smart_link.label

            # Check if the user has the smart link edit permission before
            # showing the exception text.
            try:
                AccessControlList.objects.check_access(
                    obj=self.resolved_smart_link,
                    permissions=(permission_smart_link_edit,),
                    user=self.request.user
                )
            except PermissionDenied:
                """User doesn't have the required permission."""
            else:
                messages.error(
                    message=_(
                        'Resolved smart link dynamic label error: %s' % exception
                    ), request=self.request
                )

        title = _(
            'Documents in resolved smart link "%(resolved_smart_link)s" for '
            '"%(document)s"'
        ) % {
            'document': self.external_object,
            'resolved_smart_link': resolved_smart_link_label
        }

        context = super().get_extra_context()
        context.update(
            {
                'object': self.external_object,
                'title': title
            }
        )
        return context

    def get_resolved_smart_link(self):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_resolved_smart_link_view,
            queryset=ResolvedSmartLink.objects.filter(enabled=True),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['smart_link_id']
        )


class SmartLinkListView(SingleObjectListView):
    model = SmartLink
    object_permission = permission_smart_link_view
    view_icon = icon_smart_link_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_smart_link_setup,
            'no_results_main_link': link_smart_link_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Indexes group documents into units, usually with similar '
                'properties and of equal or similar types. Smart links '
                'allow defining relationships between documents even '
                'if they are in different indexes and are of different '
                'types.'
            ),
            'no_results_title': _(
                'There are no smart links'
            ),
            'title': _('Smart links')
        }


class DocumentResolvedSmartLinkListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_permission = permission_resolved_smart_link_view
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    object_permission = permission_resolved_smart_link_view
    view_icon = icon_document_smart_link_instance_list

    def get_extra_context(self):
        return {
            'document': self.external_object,
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_smart_link_setup,
            'no_results_text': _(
                'Smart links allow defining relationships between '
                'documents even if they are in different indexes and '
                'are of different types.'
            ),
            'no_results_title': _(
                'There are no resolved smart links for this document'
            ),
            'object': self.external_object,
            'title': _(
                'Resolved smart links for document: %s'
            ) % self.external_object
        }

    def get_source_queryset(self):
        return ResolvedSmartLink.objects.get_for(
            document=self.external_object
        )


class DocumentTypeSmartLinkAddRemoveView(AddRemoveView):
    main_object_method_add_name = 'smart_links_add'
    main_object_method_remove_name = 'smart_links_remove'
    main_object_model = DocumentType
    main_object_permission = permission_document_type_edit
    main_object_pk_url_kwarg = 'document_type_id'
    secondary_object_model = SmartLink
    secondary_object_permission = permission_smart_link_edit
    list_available_title = _('Available smart links')
    list_added_title = _('Smart links enabled')
    related_field = 'smart_links'
    view_icon = icon_document_type_smart_links

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _(
                'Smart links to enable for document type: %s'
            ) % self.main_object
        }


class SmartLinkDocumentTypeAddRemoveView(AddRemoveView):
    main_object_method_add_name = 'document_types_add'
    main_object_method_remove_name = 'document_types_remove'
    main_object_permission = permission_smart_link_edit
    main_object_model = SmartLink
    main_object_pk_url_kwarg = 'smart_link_id'
    secondary_object_model = DocumentType
    secondary_object_permission = permission_document_type_edit
    list_available_title = _('Available document types')
    list_added_title = _('Document types enabled')
    related_field = 'document_types'
    view_icon = icon_smart_link_document_type_list

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _(
                'Document type for which to enable smart link: %s'
            ) % self.main_object
        }


class SmartLinkCreateView(SingleObjectCreateView):
    extra_context = {'title': _('Create new smart link')}
    form_class = SmartLinkForm
    post_action_redirect = reverse_lazy(
        viewname='linking:smart_link_list'
    )
    view_icon = icon_smart_link_create
    view_permission = permission_smart_link_create

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class SmartLinkDeleteView(SingleObjectDeleteView):
    model = SmartLink
    object_permission = permission_smart_link_delete
    pk_url_kwarg = 'smart_link_id'
    post_action_redirect = reverse_lazy(
        viewname='linking:smart_link_list'
    )
    view_icon = icon_smart_link_delete

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Delete smart link: %s') % self.object
        }


class SmartLinkEditView(SingleObjectEditView):
    form_class = SmartLinkForm
    model = SmartLink
    object_permission = permission_smart_link_edit
    pk_url_kwarg = 'smart_link_id'
    post_action_redirect = reverse_lazy(
        viewname='linking:smart_link_list'
    )
    view_icon = icon_smart_link_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Edit smart link: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}



import json

from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_workflow_instance_created, event_workflow_instance_transitioned
)
from ..permissions import (
    permission_workflow_instance_transition,
    permission_workflow_template_view, permission_workflow_tools
)

from .literals import TEST_WORKFLOW_INSTANCE_LOG_ENTRY_EXTRA_DATA
from .mixins.workflow_instance_mixins import (
    WorkflowInstanceAPIViewTestMixin, WorkflowInstanceLaunchAPIViewTestMixin,
    WorkflowInstanceLogEntryTransitionListAPIViewTestMixin
)


class WorkflowInstanceAPIViewTestCase(
    WorkflowInstanceAPIViewTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_document_stub()

    def test_workflow_instance_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_instance_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_detail_api_view_with_workflow_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_detail_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_detail_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self._test_document.workflows.first().pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_detail_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_list_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_list_api_view_with_workflow_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_document.workflows.first().pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_create_api_view_no_permission(self):
        workflow_instance = self._test_document.workflows.first()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(workflow_instance.log_entries.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_create_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(workflow_instance.log_entries.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_create_api_view_with_transition_access(self):
        self.grant_access(
            obj=self._test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(workflow_instance.log_entries.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_create_api_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self._test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(workflow_instance.log_entries.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].verb, event_workflow_instance_transitioned.id
        )

    def test_trashed_document_workflow_instance_log_entries_create_api_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self._test_workflow_template_transition,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(workflow_instance.log_entries.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_create_api_view_with_workflow_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(workflow_instance.log_entries.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_create_api_view_with_document_and_workflow_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(workflow_instance.log_entries.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].verb, event_workflow_instance_transitioned.id
        )

    def test_trashed_document_workflow_instance_log_entries_create_api_view_with_document_and_workflow_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(workflow_instance.log_entries.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_create_api_view_with_extra_data_document_and_workflow_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_instance_transition
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )
        workflow_instance = self._test_document.workflows.first()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_create_api_view(
            extra_data={
                'extra_data': TEST_WORKFLOW_INSTANCE_LOG_ENTRY_EXTRA_DATA
            }, workflow_instance=workflow_instance
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(workflow_instance.log_entries.count(), 1)
        workflow_instance.refresh_from_db()

        self.assertEqual(
            workflow_instance.get_context()['workflow_instance_context'],
            json.loads(s=TEST_WORKFLOW_INSTANCE_LOG_ENTRY_EXTRA_DATA)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].verb, event_workflow_instance_transitioned.id
        )

    def test_workflow_instance_log_entries_list_api_view_no_permission(self):
        self._create_test_workflow_template_instance_log_entry()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_list_api_view_with_document_access(self):
        self._create_test_workflow_template_instance_log_entry()

        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_list_api_view_with_workflow_access(self):
        self._create_test_workflow_template_instance_log_entry()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entries_list_api_view_with_full_access(self):
        self._create_test_workflow_template_instance_log_entry()

        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_log_entries_list_api_view_with_full_access(self):
        self._create_test_workflow_template_instance_log_entry()

        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class WorkflowInstanceLaunchAPIViewTestCase(
    WorkflowInstanceLaunchAPIViewTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(
            add_test_document_type=True, auto_launch=False
        )
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_document_stub()

    def test_workflow_instance_api_view_no_permission(self):
        test_document_workflow_instance_count = self._test_document.workflows.count()

        self._clear_events()

        response = self._request_test_workflow_instance_launch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.workflows.count(),
            test_document_workflow_instance_count
        )
        self.assertNotEqual(
            list(
                self._test_document.workflows.values_list(
                    'workflow', flat=True
                )
            ), [self._test_workflow_template.pk]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_workflow_tools
        )

        test_document_workflow_instance_count = self._test_document.workflows.count()

        self._clear_events()

        response = self._request_test_workflow_instance_launch_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self._test_document.workflows.count(),
            test_document_workflow_instance_count
        )
        self.assertNotEqual(
            list(
                self._test_document.workflows.values_list(
                    'workflow', flat=True
                )
            ), [self._test_workflow_template.pk]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_api_view_with_workflow_template_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_tools
        )

        test_document_workflow_instance_count = self._test_document.workflows.count()

        self._clear_events()

        response = self._request_test_workflow_instance_launch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.workflows.count(),
            test_document_workflow_instance_count
        )
        self.assertNotEqual(
            list(
                self._test_document.workflows.values_list(
                    'workflow', flat=True
                )
            ), [self._test_workflow_template.pk]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_workflow_tools
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_tools
        )

        test_document_workflow_instance_count = self._test_document.workflows.count()

        self._clear_events()

        response = self._request_test_workflow_instance_launch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self._test_document.workflows.count(),
            test_document_workflow_instance_count + 1
        )
        self.assertEqual(
            list(
                self._test_document.workflows.values_list(
                    'workflow', flat=True
                )
            ), [self._test_workflow_template.pk]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(
            events[0].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(events[0].verb, event_workflow_instance_created.id)


class WorkflowInstanceLogEntryTransitrionListAPIViewTestCase(
    WorkflowInstanceAPIViewTestMixin,
    WorkflowInstanceLogEntryTransitionListAPIViewTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_document_stub()

    def test_workflow_instance_log_entry_transition_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entry_transition_list_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entry_transition_list_api_view_with_template_view_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entry_transition_list_api_view_with_document_and_template_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entry_transition_list_api_view_with_transition_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entry_transition_list_api_view_with_document_and_transition_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entry_transition_list_api_view_with_teamplate_and_transition_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_instance_log_entry_transition_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 1
        )
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_workflow_template_transition.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_instance_log_entry_transition_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_instance_transition
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_instance_log_entry_transition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import migrations, models
import django.db.models.deletion


def code_delete_document_version_page_ocr_errors(apps, schema_editor):
    DocumentVersionOCRError = apps.get_model(
        app_label='ocr', model_name='DocumentVersionOCRError'
    )
    DocumentVersionOCRError.objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0075_delete_duplicateddocumentold'),
        ('ocr', '0009_auto_20210304_0950'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='DocumentTypeSettings',
            new_name='DocumentTypeOCRSettings',
        ),
        migrations.RunPython(
            code=code_delete_document_version_page_ocr_errors,
            reverse_code=migrations.RunPython.noop
        ),
        migrations.AlterField(
            model_name='documentversionocrerror',
            name='document_version',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='ocr_errors', to='documents.DocumentVersion',
                verbose_name='Document version'
            ),
        ),
        migrations.DeleteModel(
            name='DocumentPageOCRContent',
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_FILE_METADATA_AUTO_PROCESS,
    DEFAULT_FILE_METADATA_DRIVERS_ARGUMENTS
)
from .setting_migrations import FileMetadataSettingMigration

namespace = SettingNamespace(
    label=_('File metadata'), migration_class=FileMetadataSettingMigration,
    name='file_metadata', version='0002'
)
setting_auto_process = namespace.add_setting(
    choices=('false', 'true'),
    default=DEFAULT_FILE_METADATA_AUTO_PROCESS,
    global_name='FILE_METADATA_AUTO_PROCESS', help_text=_(
        'Set new document types to perform file metadata processing '
        'automatically by default.'
    )
)
setting_drivers_arguments = namespace.add_setting(
    default=DEFAULT_FILE_METADATA_DRIVERS_ARGUMENTS,
    global_name='FILE_METADATA_DRIVERS_ARGUMENTS', help_text=_(
        'Arguments to pass to the drivers.'
    )
)



import hashlib

from django.core.cache import caches
from django.utils.encoding import force_bytes

from .settings import (
    setting_document_lookup_cache_timeout, setting_node_lookup_cache_timeout
)


class MirrorFilesystemCache:
    @staticmethod
    def get_key_hash(key):
        return hashlib.sha256(
            string=force_bytes(s=key)
        ).hexdigest()

    @staticmethod
    def get_document_key(document):
        return MirrorFilesystemCache.get_key_hash(
            key='document_pk_{}'.format(document.pk)
        )

    @staticmethod
    def get_node_key(node):
        return MirrorFilesystemCache.get_key_hash(
            key='node_pk_{}'.format(node.pk)
        )

    @staticmethod
    def get_path_key(path):
        return MirrorFilesystemCache.get_key_hash(
            key='path_{}'.format(path)
        )

    def __init__(self, name='default'):
        self.cache = caches[name]

    def clear_all(self):
        self.cache.clear()

    def clear_node(self, node):
        node_key = MirrorFilesystemCache.get_node_key(node=node)
        path_cache = self.cache.get(key=node_key)
        if path_cache:
            path = path_cache.get('path')
            if path:
                self.clean_path(path=path)

        self.cache.delete(key=node_key)

    def clear_document(self, document):
        document_key = MirrorFilesystemCache.get_document_key(
            document=document
        )
        path_cache = self.cache.get(key=document_key)
        if path_cache:
            path = path_cache.get('path')
            if path:
                self.clean_path(path=path)

        self.cache.delete(key=document_key)

    def clean_path(self, path):
        self.cache.delete(
            key=MirrorFilesystemCache.get_path_key(path=path)
        )

    def get_path(self, path):
        return self.cache.get(
            key=MirrorFilesystemCache.get_path_key(path=path)
        )

    def set_path(self, path, document=None, node=None):
        # Must provide a document_pk or a node_pk
        # not both.
        if document:
            self.cache.set(
                key=MirrorFilesystemCache.get_path_key(path=path),
                timeout=setting_document_lookup_cache_timeout.value,
                value={'document_pk': document.pk}
            )
            self.cache.set(
                key=MirrorFilesystemCache.get_document_key(document=document),
                timeout=setting_document_lookup_cache_timeout.value,
                value={'path': path}
            )
        elif node:
            self.cache.set(
                key=MirrorFilesystemCache.get_path_key(path=path),
                timeout=setting_node_lookup_cache_timeout.value,
                value={'node_pk': node.pk}
            )
            self.cache.set(
                key=MirrorFilesystemCache.get_node_key(node=node),
                timeout=setting_node_lookup_cache_timeout.value,
                value={'path': path}
            )



import logging

from django import forms
from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.layers import layer_decorations
from mayan.apps.converter.transformations import (
    BaseTransformation, ImagePasteCoordinatesPercentTransformationMixin
)

logger = logging.getLogger(name=__name__)


class SignatureCapturePasteTransformation(
    ImagePasteCoordinatesPercentTransformationMixin, BaseTransformation
):
    label = _('Paste a signature capture (percent coordinates)')
    name = 'paste_signature_capture_percent'

    @staticmethod
    def get_document(page_object):
        try:
            return page_object.document_file.document
        except AttributeError:
            return page_object.document_version.document

    @classmethod
    def get_arguments(cls):
        arguments = super().get_arguments() + ('internal_name',)
        return arguments

    @classmethod
    def get_form_class(cls):
        SuperForm = super().get_form_class()

        class Form(SuperForm):
            internal_name = forms.ChoiceField(
                help_text=_('Signature capture internal name'),
                label=_('Internal name'),
                required=True
            )

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                def get_choices():
                    try:
                        page_object = self.view.external_object
                    except AttributeError:
                        page_object = self.view.object.object_layer.content_object

                    document = SignatureCapturePasteTransformation.get_document(
                        page_object=page_object
                    )
                    queryset = apps.get_model(
                        app_label='signature_captures',
                        model_name='SignatureCapture'
                    ).valid.filter(document=document)

                    for instance in queryset.all():
                        yield (instance.internal_name, instance)

                self.fields['internal_name'].choices = get_choices()

        return Form

    def get_model_instance(self):
        SignatureCapture = apps.get_model(
            app_label='signature_captures', model_name='SignatureCapture'
        )

        document = SignatureCapturePasteTransformation.get_document(
            page_object=self.object_layer.content_object
        )

        try:
            kwargs = {
                'document': document, 'internal_name': self.internal_name
            }
            instance = SignatureCapture.objects.get(**kwargs)
        except SignatureCapture.DoesNotExist:
            logger.error('Signature capture with not found.; %s', kwargs)
            raise
        else:
            return instance


BaseTransformation.register(
    layer=layer_decorations,
    transformation=SignatureCapturePasteTransformation
)



from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from django.utils.translation import ungettext, ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.organizations.utils import get_organization_installation_url
from mayan.apps.views.generics import MultipleObjectFormActionView

from ..forms import ObjectMailForm
from ..literals import MODEL_SEND_FUNCTION_DOTTED_PATH
from ..models import UserMailer
from ..permissions import permission_user_mailer_use
from ..tasks import task_send_object


class ObjectLinkMailView(MultipleObjectFormActionView):
    as_attachment = False
    form_class = ObjectMailForm

    def get_extra_context(self):
        queryset = self.object_list

        result = {
            'title': ungettext(
                singular=self.title,
                plural=self.title_plural,
                number=queryset.count()
            )
        }

        if queryset.count() == 1:
            result.update(
                {
                    'object': queryset.first(),
                    'title': _(self.title_document) % queryset.first()
                }
            )

        return result

    def get_form_extra_kwargs(self):
        return {
            'as_attachment': self.as_attachment,
            'user': self.request.user
        }

    def object_action(self, form, instance):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_user_mailer_use,
            queryset=UserMailer.objects.filter(enabled=True),
            user=self.request.user
        )

        mailer = get_object_or_404(
            klass=queryset, pk=form.cleaned_data['user_mailer'].pk
        )

        content_type = ContentType.objects.get_for_model(model=instance)

        kwargs = {
            'as_attachment': self.as_attachment,
            'body': form.cleaned_data['body'],
            'content_type_id': content_type.pk,
            'object_id': instance.pk,
            'object_name': _('Document file'),
            'organization_installation_url': get_organization_installation_url(
                request=self.request
            ),
            'recipient': form.cleaned_data['email'],
            'sender': self.request.user.email,
            'subject': form.cleaned_data['subject'],
            'user_mailer_id': mailer.pk,
            'user_id': self.request.user.pk
        }

        kwargs.update(
            MODEL_SEND_FUNCTION_DOTTED_PATH.get(
                instance._meta.model, {}
            )
        )

        task_send_object.apply_async(kwargs=kwargs)


class ObjectAttachmentMailView(ObjectLinkMailView):
    as_attachment = True



from django.conf.urls import url

from .api_views import (
    APIPermissionList, APIRoleDetailView, APIRoleGroupAddView,
    APIRoleGroupListView, APIRoleGroupRemoveView, APIRoleListView,
    APIRolePermissionAddView, APIRolePermissionListView,
    APIRolePermissionRemoveView
)

from .views import (
    GroupRoleAddRemoveView, RoleCreateView, RoleDeleteView, RoleEditView,
    RoleListView, RoleGroupAddRemoveView, RolePermissionAddRemoveView,
    StoredPermissionDetailView
)

urlpatterns = [
    url(
        regex=r'^groups/(?P<group_id>\d+)/roles/$', name='group_role_list',
        view=GroupRoleAddRemoveView.as_view()
    ),
    url(regex=r'^roles/$', name='role_list', view=RoleListView.as_view()),
    url(
        regex=r'^roles/create/$', name='role_create',
        view=RoleCreateView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>\d+)/delete/$', name='role_single_delete',
        view=RoleDeleteView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>\d+)/edit/$', name='role_edit',
        view=RoleEditView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>\d+)/groups/$', name='role_group_list',
        view=RoleGroupAddRemoveView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>\d+)/permissions/$',
        name='role_permission_list',
        view=RolePermissionAddRemoveView.as_view()
    ),
    url(
        regex=r'^roles/multiple/delete/$', name='role_multiple_delete',
        view=RoleDeleteView.as_view()
    ),
    url(
        regex=r'^stored_permissions/(?P<stored_permission_id>\d+)/$',
        name='stored_permission_detail',
        view=StoredPermissionDetailView.as_view()
    )
]

api_urls = [
    url(
        regex=r'^permissions/$', name='permission-list',
        view=APIPermissionList.as_view()
    ),
    url(regex=r'^roles/$', name='role-list', view=APIRoleListView.as_view()),
    url(
        regex=r'^roles/(?P<role_id>[0-9]+)/$', name='role-detail',
        view=APIRoleDetailView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>[0-9]+)/groups/$', name='role-group-list',
        view=APIRoleGroupListView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>[0-9]+)/groups/add/$',
        name='role-group-add', view=APIRoleGroupAddView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>[0-9]+)/groups/remove/$',
        name='role-group-remove', view=APIRoleGroupRemoveView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>[0-9]+)/permissions/$',
        name='role-permission-list', view=APIRolePermissionListView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>[0-9]+)/permissions/add/$',
        name='role-permission-add', view=APIRolePermissionAddView.as_view()
    ),
    url(
        regex=r'^roles/(?P<role_id>[0-9]+)/permissions/remove/$',
        name='role-permission-remove',
        view=APIRolePermissionRemoveView.as_view()
    )
]



from mayan.apps.appearance.classes import Icon

icon_global_error_log_entry_list = Icon(
    driver_name='fontawesome', symbol='exclamation-triangle'
)
icon_object_errors = Icon(
    driver_name='fontawesome', symbol='exclamation-triangle'
)
icon_object_error_log_entry_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_object_error_log_entry_list = Icon(
    driver_name='fontawesome', symbol='exclamation-triangle'
)
icon_object_error_log_entry_list_clear = Icon(
    driver_name='fontawesome-dual', primary_symbol='exclamation-triangle',
    secondary_symbol='times'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Quotas'), name='quotas'
)

permission_quota_create = namespace.add_permission(
    label=_('Create a quota'), name='quota_create'
)
permission_quota_delete = namespace.add_permission(
    label=_('Delete a quota'), name='quota_delete'
)
permission_quota_edit = namespace.add_permission(
    label=_('Edit a quota'), name='quota_edit'
)
permission_quota_view = namespace.add_permission(
    label=_('View a quota'), name='quota_view'
)



import json
import logging

from django.db import models
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from .literals import IMPORT_ERROR_EXCLUSION_TEXTS

logger = logging.getLogger(name=__name__)


class BackendModelMixin(models.Model):
    """
    Backends here represent drivers. This model allows storing multiple
    instances of a single backend.
    """
    _backend_model_null_backend = None

    backend_path = models.CharField(
        max_length=128, help_text=_(
            'The dotted Python path to the backend class.'
        ), verbose_name=_('Backend path')
    )
    backend_data = models.TextField(
        blank=True, help_text=_(
            'JSON encoded data for the backend class.'
        ), verbose_name=_('Backend data')
    )

    class Meta:
        abstract = True

    def get_backend_class(self):
        """
        Retrieves the backend by importing the module and the class.
        """
        try:
            return import_string(dotted_path=self.backend_path)
        except ModuleNotFoundError as exception:
            logger.error(
                'ModuleNotFoundError while importing backend: %s; %s',
                self.backend_path, exception
            )
            if self._backend_model_null_backend:
                return self._backend_model_null_backend
            else:
                raise
        except ImportError as exception:
            logger.error(
                'ImportError while importing backend: %s; %s',
                self.backend_path, exception
            )
            for import_error_exclusion_text in IMPORT_ERROR_EXCLUSION_TEXTS:
                if import_error_exclusion_text in str(exception):
                    raise

            if self._backend_model_null_backend:
                return self._backend_model_null_backend
            else:
                raise

    def get_backend_class_label(self):
        """
        Return the label that the backend itself provides. The backend is
        loaded but not initialized. As such the label returned is a class
        property.
        """
        backend_class = self.get_backend_class()

        return backend_class.label

    get_backend_class_label.short_description = _('Backend')
    get_backend_class_label.help_text = _('The backend class for this entry.')

    def get_backend_data(self):
        return json.loads(s=self.backend_data or '{}')

    def set_backend_data(self, obj):
        self.backend_data = json.dumps(obj=obj)

    def get_backend_instance(self):
        backend_class = self.get_backend_class()

        kwargs = self.get_backend_data()

        return backend_class(model_instance_id=self.id, **kwargs)



import json

from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import StoredCredential

from .literals import (
    TEST_STORED_CREDENTIAL_BACKEND_DATA_FIELDS,
    TEST_STORED_CREDENTIAL_BACKEND_DATA_PASSWORD,
    TEST_STORED_CREDENTIAL_BACKEND_DATA_USERNAME,
    TEST_STORED_CREDENTIAL_BACKEND_PATH,
    TEST_STORED_CREDENTIAL_INTERNAL_NAME, TEST_STORED_CREDENTIAL_LABEL,
    TEST_STORED_CREDENTIAL_LABEL_EDITED
)


class StoredCredentialTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = StoredCredential
    _test_object_name = '_test_stored_credential'
    _test_stored_credential_backend_data = TEST_STORED_CREDENTIAL_BACKEND_DATA_FIELDS
    _test_stored_credential_backend_path = TEST_STORED_CREDENTIAL_BACKEND_PATH
    auto_create_test_credential = True

    def setUp(self):
        super().setUp()
        self._test_stored_credential_list = []

        if self.auto_create_test_credential:
            self._create_test_stored_credential()

    def _create_test_stored_credential(self, backend_data=None):
        total_test_stored_credential_count = len(
            self._test_stored_credential_list
        )
        label = '{}_{}'.format(
            TEST_STORED_CREDENTIAL_LABEL, total_test_stored_credential_count
        )

        if backend_data is None:
            backend_data = self._test_stored_credential_backend_data

        backend_data_serialized = json.dumps(
            obj=backend_data
        )

        self._test_stored_credential = StoredCredential.objects.create(
            backend_data=backend_data_serialized,
            backend_path=self._test_stored_credential_backend_path,
            internal_name=TEST_STORED_CREDENTIAL_INTERNAL_NAME, label=label
        )
        self._test_stored_credential_list.append(
            self._test_stored_credential
        )


class StoredCredentialAPIViewTestMixin(StoredCredentialTestMixin):
    def _request_test_stored_credential_create_api_view(self):
        self._test_object_track()

        backend_data = json.dumps(
            obj=TEST_STORED_CREDENTIAL_BACKEND_DATA_FIELDS
        )

        response = self.post(
            viewname='rest_api:credential-list', data={
                'internal_name': TEST_STORED_CREDENTIAL_INTERNAL_NAME,
                'label': TEST_STORED_CREDENTIAL_LABEL,
                'backend_path': TEST_STORED_CREDENTIAL_BACKEND_PATH,
                'backend_data': backend_data
            }
        )

        self._test_object_set()

        return response

    def _request_test_stored_credential_delete_api_view(self):
        return self.delete(
            viewname='rest_api:credential-detail', kwargs={
                'credential_id': self._test_stored_credential.pk
            }
        )

    def _request_test_stored_credential_detail_api_view(self):
        return self.get(
            viewname='rest_api:credential-detail', kwargs={
                'credential_id': self._test_stored_credential.pk
            }
        )

    def _request_test_stored_credential_edit_via_patch_api_view(self):
        return self.patch(
            viewname='rest_api:credential-detail', kwargs={
                'credential_id': self._test_stored_credential.pk
            }, data={
                'label': TEST_STORED_CREDENTIAL_LABEL_EDITED
            }
        )

    def _request_test_stored_credential_list_api_view(self):
        return self.get(viewname='rest_api:credential-list')


class StoredCredentialPasswordUsernameTestMixin(StoredCredentialTestMixin):
    _test_stored_credential_backend_data = {
        'password': TEST_STORED_CREDENTIAL_BACKEND_DATA_PASSWORD,
        'username': TEST_STORED_CREDENTIAL_BACKEND_DATA_USERNAME
    }
    _test_stored_credential_backend_path = 'mayan.apps.credentials.credential_backends.CredentialBackendUsernamePassword'

    def _create_test_stored_credential(self, *args, **kwargs):
        super()._create_test_stored_credential(*args, **kwargs)

        self._test_stored_credential._password = self._test_stored_credential_backend_data['password']
        self._test_stored_credential._username = self._test_stored_credential_backend_data['username']


class StoredCredentialViewTestMixin(StoredCredentialTestMixin):
    def _request_test_stored_credential_backend_selection_view(self):
        return self.post(
            viewname='credentials:stored_credential_backend_selection',
            data={
                'backend': TEST_STORED_CREDENTIAL_BACKEND_PATH,
            }
        )

    def _request_test_stored_credential_create_view(self):
        self._test_object_track()

        backend_data = json.dumps(
            obj=TEST_STORED_CREDENTIAL_BACKEND_DATA_FIELDS
        )

        data = {
            'backend_data': backend_data,
            'internal_name': TEST_STORED_CREDENTIAL_INTERNAL_NAME,
            'label': TEST_STORED_CREDENTIAL_LABEL
        }

        response = self.post(
            viewname='credentials:stored_credential_create', kwargs={
                'backend_path': TEST_STORED_CREDENTIAL_BACKEND_PATH,
            }, data=data
        )

        self._test_object_set()

        return response

    def _request_test_stored_credential_delete_view(self):
        return self.post(
            viewname='credentials:stored_credential_delete', kwargs={
                'stored_credential_id': self._test_stored_credential.pk
            }
        )

    def _request_test_stored_credential_edit_view(self):
        return self.post(
            viewname='credentials:stored_credential_edit', kwargs={
                'stored_credential_id': self._test_stored_credential.pk
            }, data={
                'label': TEST_STORED_CREDENTIAL_LABEL_EDITED,
                'internal_name': self._test_stored_credential.internal_name
            }
        )

    def _request_test_stored_credential_list_view(self):
        return self.get(viewname='credentials:stored_credential_list')
